#!/usr/bin/env python

"""Plotting methods.

This module contains various plotting methods.

"""

from __future__ import division

import numpy as np
from matplotlib import pyplot as plt
from scipy import interpolate

from wyrm import tentensystem as tts


def plot_scalp(v, channel):
    """Plot the values v for channel ``channel`` on a scalp."""

    channelpos = [tts.channels[c] for c in channel]
    points = [calculate_stereographic_projection(i) for i in channelpos]
    x = [i[0] for i in points]
    y = [i[1] for i in points]
    z = v
    X, Y, Z = interpolate_2d(x, y, z)
    plt.contour(X, Y, Z, 20)
    plt.contourf(X, Y, Z, 20)
    #plt.clabel(im)
    plt.colorbar()
    plt.gca().add_artist(plt.Circle((0, 0), radius=1, linewidth=3, fill=False))
    plt.plot(x, y, 'bo')
    for i in zip(channel, zip(x,y)):
        plt.annotate(i[0], i[1])


def plot_channels(dat, chanaxis=-1, otheraxis=-2):
    """Plot all channels for a continuous.

    Parameters
    ----------
    dat : Data

    """
    ax = []
    n_channels = dat.data.shape[chanaxis]
    for i, chan in enumerate(dat.axes[chanaxis]):
        if i == 0:
            a = plt.subplot(10, n_channels / 10 + 1, i + 1)
        else:
            a = plt.subplot(10, n_channels / 10 + 1, i + 1, sharex=ax[0], sharey=ax[0])
        ax.append(a)
        x, y =  dat.axes[otheraxis], dat.data.take([i], chanaxis)
        a.plot(dat.axes[otheraxis], dat.data.take([i], chanaxis).squeeze())
        a.set_title(chan)
        plt.axvline(x=0)
        plt.axhline(y=0)


def plot_spectrum(spectrum, freqs):
    plt.plot(freqs, spectrum, '.')
    plt.xlabel('Frequency [Hz]')
    plt.ylabel('[dl]')


def plot_spectrogram(spectrogram, freqs):
    extent = 0, len(spectrogram), freqs[0], freqs[-1]
    plt.imshow(spectrogram.transpose(),
        aspect='auto',
        origin='lower',
        extent=extent,
        interpolation='none')
    plt.colorbar()
    plt.ylabel('Frequency [Hz]')
    plt.xlabel('Time')


def calculate_stereographic_projection(p):
    """Calculate the stereographic projection.

    Given a unit sphere with radius ``r = 1`` and center at the origin.
    Project the point ``p = (x, y, z)`` from the sphere's South pole (0,
    0, -1) on a plane on the sphere's North pole (0, 0, 1).

    The formula is:

        P' = P * (2r / (r + z))

    Parameters
    ----------
    p : [float, float]
        The point to be projected in cartesian coordinates.

    Returns
    -------
    x, y : float, float
        The projected point on the plane.

    """
    # P' = P * (2r / r + z)
    mu = 1 / (1 + p[2])
    x = p[0] * mu
    y = p[1] * mu
    return x, y


def interpolate_2d(x, y, z):
    """Interpolate missing points on a plane.

    Parameters
    ----------
    x, y, z : equally long lists of floats
        1d arrays defining points like ``p[x, y] = z``

    Returns
    -------
    X, Y, Z : 1d array, 1d array, 2d array
        ``Z`` is a 2d array ``[min(x)..max(x), [min(y)..max(y)]`` with
        the interpolated values as values.

    """
    X = np.linspace(min(x), max(x))
    Y = np.linspace(min(y), max(y))
    X, Y = np.meshgrid(X, Y)
    #f = interpolate.interp2d(x, y, z)
    #Z = f(X[0, :], Y[:, 0])
    f = interpolate.LinearNDInterpolator(zip(x, y), z)
    Z = f(X, Y)
    return X, Y, Z



from __future__ import division

import unittest

import numpy as np

from wyrm.types import Data
from wyrm.processing import segment_dat
from wyrm.processing import swapaxes


class TestSegmentDat(unittest.TestCase):

    def setUp(self):
        # create 100 samples and tree channels data
        ones = np.ones((100, 3))
        data = np.array([ones, ones*2, ones*3]).reshape(-1, 3)
        time = np.linspace(0, 3000, 300, endpoint=False)
        channels = ['a', 'b', 'c']
        markers = [[500, 'M1'], [1500, 'M2'], [2500, 'M3']]
        self.dat = Data(data, [time, channels], ['time', 'channels'], ['ms', '#'])
        self.dat.markers = markers
        self.dat.fs = 100
        self.mrk_def = {'class 1': ['M1'],
                        'class 2': ['M2', 'M3']
                       }

    def test_segment_dat(self):
        """Test conversion from Continuous to Epoched data."""
        epo = segment_dat(self.dat, self.mrk_def, [-400, 400])
        # test if basic info was transferred from cnt
        self.assertEqual(self.dat.markers, epo.markers)
        self.assertEqual(self.dat.fs, epo.fs)
        np.testing.assert_array_equal(self.dat.axes[-1], epo.axes[-1])
        # test if the actual data is correct
        self.assertEqual(list(epo.axes[0]), [0, 1, 1])
        np.testing.assert_array_equal(epo.class_names, np.array(['class 1', 'class 2']))
        self.assertEqual(epo.data.shape, (3, 80, 3))
        for i in range(3):
            e = epo.data[i, ...]
            self.assertEqual(np.average(e), i+1)
        # test if the epo.ival is the same we cut out
        self.assertEqual(epo.axes[-2][0], -400)
        self.assertEqual(epo.axes[-2][-1], 390)

    def test_segment_dat_with_nonexisting_markers(self):
        """Segmentation without result should return empty .data"""
        mrk_def = {'class 1': ['FUU1'],
                   'class 2': ['FUU2', 'FUU3']
                  }
        epo = segment_dat(self.dat, mrk_def, [-400, 400])
        self.assertEqual(epo.data.shape[0], 0)

    def test_segment_dat_with_unequally_sized_data(self):
        """Segmentation must ignore too short or too long chunks in the result."""
        # We create a marker that is too close to the beginning of the
        # data, so its cnt will not bee of length [-400, 400] ms. It
        # should not appear in the resulting epo
        self.dat.markers.append([100, 'M1'])
        epo = segment_dat(self.dat, self.mrk_def, [-400, 400])
        self.assertEqual(epo.data.shape[0], 3)

    # the following tests
    # (test_segment_dat_with_restriction_to_new_data_ival...) work very
    # similar but test slightly different conditions. The basic idea is
    # always: we create a small cnt with three markers directly next to
    # each other, the only thing changing between the tests is the
    # interval. We test all possible combinations of the marker position
    # relative to the interval:
    #   [M----], M [---], [--M--], [----M], [---] M
    # we check in each test that with increasing number of new samples
    # the correct number of epochs is returned.
    # WARNING: This is fairly complicated to get right, if you want to
    # change something please make sure you fully understand the problem
    def test_segment_dat_with_restriction_to_new_data_ival_zero_pos(self):
        """Online Segmentation with ival 0..+something must work correctly."""
        # [   0.,  100.,  200.,  300.,  400.,  500.,  600.,  700.,  800.]
        #         M100                                600
        #                M200                                700
        #                       M299                                799
        #                       M300                                800
        #                       M301                                801
        data = np.ones((9, 3))
        time = np.linspace(0, 900, 9, endpoint=False)
        channels = 'a', 'b', 'c'
        markers = [[100, 'x'], [200, 'x'], [299, 'x'], [300, 'x'], [301, 'x']]
        dat = Data(data, [time, channels], ['time', 'channels'], ['ms', '#'])
        dat.fs = 10
        dat.markers = markers
        mrk_def = {'class 1': ['x']}
        # each tuple has (number of new samples, expected epocs)
        samples_epos = [(0, 0), (1, 1), (2, 3), (3, 4), (4, 5), (5, 5)]
        for s, e in samples_epos:
            epo = segment_dat(dat, mrk_def, [0, 500], newsamples=s)
            self.assertEqual(epo.data.shape[0], e)

    def test_segment_dat_with_restriction_to_new_data_ival_pos_pos(self):
        """Online Segmentation with ival +something..+something must work correctly."""
        # [   0.,  100.,  200.,  300.,  400.,  500.,  600.,  700.,  800.]
        #         M100    200                         600
        #                M200    300                         700
        #                       M299    399                         799
        #                       M300    400                         800
        #                       M301    401                         801
        data = np.ones((9, 3))
        time = np.linspace(0, 900, 9, endpoint=False)
        channels = 'a', 'b', 'c'
        markers = [[100, 'x'], [200, 'x'], [299, 'x'], [300, 'x'], [301, 'x']]
        dat = Data(data, [time, channels], ['time', 'channels'], ['ms', '#'])
        dat.fs = 10
        dat.markers = markers
        mrk_def = {'class 1': ['x']}
        # each tuple has (number of new samples, expected epocs)
        samples_epos = [(0, 0), (1, 1), (2, 3), (3, 4), (4, 5), (5, 5)]
        for s, e in samples_epos:
            epo = segment_dat(dat, mrk_def, [100, 500], newsamples=s)
            self.assertEqual(epo.data.shape[0], e)

    def test_segment_dat_with_restriction_to_new_data_ival_neg_pos(self):
        """Online Segmentation with ival -something..+something must work correctly."""
        # [   0.,  100.,  200.,  300.,  400.,  500.,  600.,  700.,  800.]
        #          100                 M400           600
        #                 200                 M500           700
        #                        299                 M599           799
        #                        300                 M600           800
        #                        301                 M601           801
        data = np.ones((9, 3))
        time = np.linspace(0, 900, 9, endpoint=False)
        channels = 'a', 'b', 'c'
        markers = [[400, 'x'], [500, 'x'], [599, 'x'], [600, 'x'], [601, 'x']]
        dat = Data(data, [time, channels], ['time', 'channels'], ['ms', '#'])
        dat.fs = 10
        dat.markers = markers
        mrk_def = {'class 1': ['x']}
        # each tuple has (number of new samples, expected epocs)
        samples_epos = [(0, 0), (1, 1), (2, 3), (3, 4), (4, 5), (5, 5)]
        for s, e in samples_epos:
            epo = segment_dat(dat, mrk_def, [-300, 200], newsamples=s)
            self.assertEqual(epo.data.shape[0], e)

    def test_segment_dat_with_restriction_to_new_data_ival_neg_zero(self):
        """Online Segmentation with ival -something..0 must work correctly."""
        # [   0.,  100.,  200.,  300.,  400.,  500.,  600.,  700.,  800.]
        #          100                        M500
        #                 200                        M600
        #                        299                        M699
        #                        300                        M700
        #                        301                        M701
        data = np.ones((9, 3))
        time = np.linspace(0, 900, 9, endpoint=False)
        channels = 'a', 'b', 'c'
        markers = [[500, 'x'], [600, 'x'], [699, 'x'], [700, 'x'], [701, 'x']]
        dat = Data(data, [time, channels], ['time', 'channels'], ['ms', '#'])
        dat.fs = 10
        dat.markers = markers
        mrk_def = {'class 1': ['x']}
        # each tuple has (number of new samples, expected epocs)
        samples_epos = [(0, 0), (1, 0), (2, 2), (3, 4), (4, 5), (5, 5)]
        for s, e in samples_epos:
            epo = segment_dat(dat, mrk_def, [-400, 0], newsamples=s)
            self.assertEqual(epo.data.shape[0], e)

    def test_segment_dat_with_restriction_to_new_data_ival_neg_neg(self):
        """Online Segmentation with ival -something..-something must work correctly."""
        # [   0.,  100.,  200.,  300.,  400.,  500.,  600.,  700.,  800.]
        #          100                  400   M500
        #                 200                  500   M600
        #                        299                  599   M699
        #                        300                  600   M700
        #                        301                  600   M701
        data = np.ones((9, 3))
        time = np.linspace(0, 900, 9, endpoint=False)
        channels = 'a', 'b', 'c'
        markers = [[500, 'x'], [600, 'x'], [699, 'x'], [700, 'x'], [701, 'x']]
        dat = Data(data, [time, channels], ['time', 'channels'], ['ms', '#'])
        dat.fs = 10
        dat.markers = markers
        mrk_def = {'class 1': ['x']}
        # each tuple has (number of new samples, expected epocs)
        samples_epos = [(0, 0), (1, 0), (2, 2), (3, 4), (4, 5), (5, 5)]
        for s, e in samples_epos:
            epo = segment_dat(dat, mrk_def, [-400, -100], newsamples=s)
            self.assertEqual(epo.data.shape[0], e)

    def test_segment_dat_with_negative_newsamples(self):
        """Raise an error when newsamples is not positive or None."""
        with self.assertRaises(AssertionError):
            segment_dat(self.dat, self.mrk_def, [-400, 400], newsamples=-1)

    def test_segment_dat_copy(self):
        """segment_dat must not modify arguments."""
        cpy = self.dat.copy()
        segment_dat(self.dat, self.mrk_def, [-400, 400])
        self.assertEqual(cpy, self.dat)

    def test_segment_dat_swapaxes(self):
        """Segmentation must work with nonstandard axes."""
        epo = segment_dat(swapaxes(self.dat, 0, 1), self.mrk_def, [-400, 400], timeaxis=-1)
        # segment_dat added a new dimension
        epo = swapaxes(epo, 1, 2)
        epo2 = segment_dat(self.dat, self.mrk_def, [-400, 400])
        self.assertEqual(epo, epo2)

if __name__ == '__main__':
    unittest.main()



from __future__ import division

import unittest

import numpy as np

from wyrm.types import Data
from wyrm.processing import logarithm

class TestLogarithm(unittest.TestCase):

    def setUp(self):
        raw = np.arange(1, 21).reshape(4, 5)
        channels = ['ca1', 'ca2', 'cb1', 'cb2', 'cc1']
        time = np.arange(4)
        self.dat = Data(raw, [time, channels], ['time', 'channels'], ['ms', '#'])

    def test_logarithm(self):
        """logarithm basics must work."""
        dat = logarithm(self.dat)
        # works elementwise (does not alter the shape)
        self.assertEqual(self.dat.data.shape, dat.data.shape)
        # actual log was computed
        np.testing.assert_array_almost_equal(np.e**dat.data, self.dat.data)

    def test_logarithm_copy(self):
        """Rectify channels must not change the original parameter."""
        cpy = self.dat.copy()
        logarithm(self.dat)
        self.assertEqual(cpy, self.dat)

if __name__ == '__main__':
    unittest.main()



from __future__ import division

import unittest

import numpy as np

from wyrm.plot import interpolate_2d


class TestInterpolate2d(unittest.TestCase):

    def test_max(self):
        """Make sure the interpolation does not lead to interpolated values
        bigger than the maximum provided by the points.

        This particularly happens when doing interpolation with splines.

        """
        x = [1, 2, 1, 2, 3]
        y = [1, 1, 2, 2, 3]
        z = [10, 10, 10, 10, 2]
        _, _, Z = interpolate_2d(x, y, z)
        self.assertAlmostEqual(np.nanmax(Z), 10.)


if __name__ == '__main__':
    unittest.main()



from __future__ import division

import unittest

import numpy as np

from wyrm.types import Data
from wyrm.processing import append_cnt
from wyrm.processing import swapaxes


class TestAppendCnt(unittest.TestCase):

    def setUp(self):
        ones = np.ones((10, 5))
        # cnt with 1, 2, 3
        cnt = np.append(ones, ones*2, axis=0)
        cnt = np.append(cnt, ones*3, axis=0)
        channels = ['ca1', 'ca2', 'cb1', 'cb2', 'cc1']
        time = np.linspace(0, 3000, 30, endpoint=False)
        self.dat = Data(cnt, [time, channels], ['time', 'channel'], ['ms', '#'])
        self.dat.markers = [[0, 'a'], [1, 'b']]
        self.dat.fs = 10

    def test_append_cnt(self):
        """append_cnt."""
        dat = append_cnt(self.dat, self.dat)
        self.assertEqual(dat.data.shape[0], 2*self.dat.data.shape[0])
        self.assertEqual(len(dat.axes[0]), 2*len(self.dat.axes[0]))
        np.testing.assert_array_equal(dat.data, np.concatenate([self.dat.data, self.dat.data], axis=0))
        np.testing.assert_array_equal(dat.axes[0], np.linspace(0, 6000, 60, endpoint=False))
        self.assertEqual(dat.markers, self.dat.markers + map(lambda x: [x[0] + 3000, x[1]], self.dat.markers))

    def test_append_cnt_with_extra(self):
        """append_cnt with extra must work with list and ndarrays."""
        self.dat.a = range(10)
        self.dat.b = np.arange(10)
        dat = append_cnt(self.dat, self.dat, extra=['a', 'b'])
        self.assertEqual(dat.a, range(10) + range(10))
        np.testing.assert_array_equal(dat.b, np.concatenate([np.arange(10), np.arange(10)]))
        self.assertEqual(dat.markers, self.dat.markers + map(lambda x: [x[0] + 3000, x[1]], self.dat.markers))

    def test_append_cnt_swapaxes(self):
        """append_cnt must work with nonstandard timeaxis."""
        dat = append_cnt(swapaxes(self.dat, 0, 1), swapaxes(self.dat, 0, 1), timeaxis=1)
        dat = swapaxes(dat, 0, 1)
        dat2 = append_cnt(self.dat, self.dat)
        self.assertEqual(dat, dat2)

    def test_append_cnt_copy(self):
        """append_cnt means must not modify argument."""
        cpy = self.dat.copy()
        append_cnt(self.dat, self.dat)
        self.assertEqual(self.dat, cpy)


if __name__ == '__main__':
    unittest.main()



from __future__ import division

import unittest

import numpy as np

from wyrm.types import Data
from wyrm.processing import variance
from wyrm.processing import swapaxes


class TestVariance(unittest.TestCase):

    def setUp(self):
        ones = np.ones((10, 5))
        # epo with 0, 1, 2
        data = np.array([0*ones, ones, 2*ones])
        channels = ['ca1', 'ca2', 'cb1', 'cb2', 'cc1']
        time = np.linspace(0, 1000, 10, endpoint=False)
        classes = [0, 1, 2]
        self.dat = Data(data, [classes, time, channels], ['class', 'time', 'channel'], ['#', 'ms', '#'])

    def test_variance(self):
        """Variance."""
        dat = variance(self.dat)
        # test the resulting dat has one axis less (the middle one)
        self.assertEqual(dat.data.shape, self.dat.data.shape[::2])
        # each epoch should have a variance of zero, test if the var of
        # all epochs is 0
        self.assertEqual(dat.data.var(), 0)

    def test_variance_with_cnt(self):
        """variance must work with cnt argument."""
        data = self.dat.data[1]
        axes = self.dat.axes[1:]
        names = self.dat.names[1:]
        units = self.dat.units[1:]
        dat = self.dat.copy(data=data, axes=axes, names=names, units=units)
        dat = variance(dat)
        self.assertEqual(dat.data.var(), 0)

    def test_variance_swapaxes(self):
        """variance must work with nonstandard timeaxis."""
        dat = variance(swapaxes(self.dat, 1, 2), timeaxis=2)
        # we don't swap back here as variance removes the timeaxis
        dat2 = variance(self.dat)
        self.assertEqual(dat, dat2)

    def test_variance_copy(self):
        """variance must not modify argument."""
        cpy = self.dat.copy()
        variance(self.dat)
        self.assertEqual(self.dat, cpy)


if __name__ == '__main__':
    unittest.main()



from __future__ import division

import unittest

import numpy as np

from wyrm.types import Data
from wyrm.processing import calculate_signed_r_square
from wyrm.processing import swapaxes


class TestSelectClasses(unittest.TestCase):

    def setUp(self):
        # create noisy data [40 epochs, 100 samples, 64 channels] with
        # values 0..1
        dat = np.random.uniform(size=(40, 100, 64))
        # every second epoch belongs to class 0 and 1 alterning
        # for class 1 add 1 in interval 40..80
        # for class 2 add 1 in interval 20..60
        #
        #        * .* .
        #
        # .* .* .      * .*
        # ------------------>
        # 0  20 40 60 80 100
        dat[::2,40:80:,:] += 1
        dat[1::2,20:60:,:] += 1
        time = np.arange(dat.shape[1])
        classes = np.zeros(dat.shape[0])
        classes[::2] = 1
        chans = np.arange(64)
        self.dat = Data(dat, [classes, time, chans], ['class', 'time', 'channel'], ['#', 'ms', '#'])
        self.dat.class_names = 'one', 'two'

    def test_calculate_signed_r_square(self):
        """Calculating signed r**2."""
        dat = calculate_signed_r_square(self.dat)
        self.assertEqual(dat.ndim + 1, self.dat.data.ndim)
        # average over channels (one could also take just one channel)
        dat = dat.mean(axis=1)
        # check the intervals
        self.assertTrue(all(dat[0:20] < .2))
        self.assertTrue(all(dat[20:40] > .5))
        self.assertTrue(all(dat[40:60] < .2))
        self.assertTrue(all(dat[60:80] < .5))
        self.assertTrue(all(dat[80:100] < .2))

    def test_calculate_signed_r_square_min_max(self):
        """Min and max values must be in [-1, 1]."""
        dat = calculate_signed_r_square(self.dat)
        self.assertTrue(-1 <= np.min(dat) <= 1)
        self.assertTrue(-1 <= np.max(dat) <= 1)

    def test_calculate_signed_r_square_with_cnt(self):
        """Select epochs must raise an exception if called with cnt argument."""
        del(self.dat.class_names)
        with self.assertRaises(AssertionError):
            calculate_signed_r_square(self.dat)

    def test_calculate_signed_r_square_swapaxes(self):
        """caluclate_r_square must work with nonstandard classaxis."""
        dat = calculate_signed_r_square(swapaxes(self.dat, 0, 2), classaxis=2)
        # the class-axis just dissapears during
        # calculate_signed_r_square, so axis 2 becomes axis 1
        dat = dat.swapaxes(0, 1)
        dat2 = calculate_signed_r_square(self.dat)
        np.testing.assert_array_equal(dat, dat2)

    def test_calculate_signed_r_square_copy(self):
        """caluclate_r_square must not modify argument."""
        cpy = self.dat.copy()
        calculate_signed_r_square(self.dat)
        self.assertEqual(self.dat, cpy)


if __name__ == '__main__':
    unittest.main()



from __future__ import division

import unittest

import numpy as np

from wyrm.types import Data
from wyrm.processing import calculate_classwise_average


class TestCalculateClasswiseAverage(unittest.TestCase):

    def setUp(self):
        ones = np.ones((10, 2))
        twoes = ones * 2
        # 7 epochs
        data = np.array([ones, ones, twoes, twoes, ones, twoes, twoes])
        channels = ['c1', 'c2']
        time = np.linspace(0, 1000, 10)
        classes = [0, 0, 1, 1, 0, 1, 1]
        class_names = ['ones', 'twoes']
        self.dat = Data(data, [classes, time, channels], ['class', 'time', 'channel'], ['#', 'ms', '#'])
        self.dat.class_names = class_names

    def test_calculate_classwise_average(self):
        """Calculate classwise average."""
        avg_dat = calculate_classwise_average(self.dat)
        # check for two datches (one for each class)
        self.assertEqual(avg_dat.data.shape[0], 2)
        # check if the data is correct
        self.assertEqual(np.average(avg_dat.data[0]), 1)
        self.assertEqual(np.average(avg_dat.data[1]), 2)
        # check if we have as many classes on axes as we have in data
        self.assertEqual(avg_dat.data.shape[0], len(avg_dat.axes[0]))
        #
        self.assertEqual(avg_dat.class_names, self.dat.class_names)

    def test_calculate_classwise_average_with_cnt(self):
        """Calculate classwise avg must raise an error if called with continouos data."""
        del(self.dat.class_names)
        with self.assertRaises(AssertionError):
            calculate_classwise_average(self.dat)

    def test_calculate_classwise_average_copy(self):
        """Calculate classwise avg must not modify the argument."""
        cpy = self.dat.copy()
        calculate_classwise_average(self.dat)
        self.assertEqual(self.dat, cpy)

if __name__ == '__main__':
    unittest.main()



from __future__ import division

import unittest

import numpy as np

from wyrm.types import Data
from wyrm.processing import subsample
from wyrm.processing import swapaxes

class TestSubsample(unittest.TestCase):

    def setUp(self):
        raw = np.arange(2000).reshape(-1, 5)
        channels = ['ca1', 'ca2', 'cb1', 'cb2', 'cc1']
        time = np.linspace(0, 4000, 400, endpoint=False)
        fs = 100
        marker = [[100, 'foo'], [200, 'bar']]
        self.dat = Data(raw, [time, channels], ['time', 'channels'], ['ms', '#'])
        self.dat.fs = fs
        self.dat.markers = marker

    def test_subsampling(self):
        """Subsampling to 10Hz."""
        dat = subsample(self.dat, 10)
        # check if the new fs is correct
        self.assertEqual(dat.fs, 10.)
        # check if data and axes have the same length
        self.assertEqual(len(dat.axes[-1]), dat.data.shape[-1])
        # no channels must have been deleted
        np.testing.assert_array_equal(self.dat.axes[-1], dat.axes[-1])
        self.assertEqual(self.dat.data.shape[-1], dat.data.shape[-1])
        # markers must not have been modified
        self.assertEqual(self.dat.markers, dat.markers)
        # no marker must have been deleted
        self.assertEqual(len(self.dat.markers), len(dat.markers))
        # check the actual data
        # after subsampling, data should look like:
        # [[0,   1,  2,  3,  4,  5]
        #  [50, 51, 52, 53, 54, 55]
        #  [...]]
        # so the first column of the resampled data should be all
        # multiples of 50.
        zeros = dat.data[:, 0] % 50
        self.assertFalse(np.any(zeros))

    def test_subsample_with_epo(self):
        """subsample must work with epoched data."""
        data = np.array([self.dat.data, self.dat.data, self.dat.data])
        axes = [np.arange(3), self.dat.axes[0], self.dat.axes[1]]
        names = ['class', 'time', 'channel']
        units = ['#', 'ms', '#']
        dat = self.dat.copy(data=data, axes=axes, names=names, units=units)
        dat = subsample(dat, 10)
        self.assertEqual(dat.fs, 10)
        self.assertEqual(dat.data.ndim, 3)
        self.assertEqual(len(dat.axes[1]), dat.data.shape[1])
        self.assertEqual(dat.data.shape[1], self.dat.data.shape[0] / 10)

    def test_whole_number_divisor_check(self):
        """Freq must be a whole number divisor of dat.fs"""
        with self.assertRaises(AssertionError):
            subsample(self.dat, 33)
        with self.assertRaises(AssertionError):
            subsample(self.dat, 101)

    def test_has_fs_check(self):
        """subsample must raise an exception if .fs attribute is not found."""
        with self.assertRaises(AssertionError):
            del(self.dat.fs)
            subsample(self.dat, 10)

    def test_axes_and_data_have_same_len_check(self):
        """subsample must raise an error if the timeaxis and data have not the same lengh."""
        with self.assertRaises(AssertionError):
            self.dat.axes[-2] = self.dat.axes[-2][1:]
            subsample(self.dat, 10)

    def test_subsample_copy(self):
        """Subsample must not modify argument."""
        cpy = self.dat.copy()
        subsample(self.dat, 10)
        self.assertEqual(cpy, self.dat)

    def test_subsample_swapaxes(self):
        """subsample must work with nonstandard timeaxis."""
        dat = subsample(swapaxes(self.dat, 0, 1), 10, timeaxis=1)
        dat = swapaxes(dat, 0, 1)
        dat2 = subsample(self.dat, 10)
        self.assertEqual(dat, dat2)


if __name__ == '__main__':
    unittest.main()



from __future__ import division

import unittest

import numpy as np

from wyrm.types import Data
from wyrm.processing import jumping_means
from wyrm.processing import swapaxes


class TestJumpingMeans(unittest.TestCase):

    def setUp(self):
        ones = np.ones((10, 5))
        # cnt with 1, 2, 3
        cnt = np.append(ones, ones*2, axis=0)
        cnt = np.append(cnt, ones*3, axis=0)
        channels = ['ca1', 'ca2', 'cb1', 'cb2', 'cc1']
        time = np.linspace(0, 3000, 30, endpoint=False)
        classes = [0, 1, 2, 1]
        # four cnts: 1s, -1s, and 0s
        data = np.array([cnt * 0, cnt * 1, cnt * 2, cnt * 0])
        self.dat = Data(data, [classes, time, channels], ['class', 'time', 'channel'], ['#', 'ms', '#'])

    def test_jumping_means(self):
        """Jumping means."""
        # with several ivals
        dat = jumping_means(self.dat, [[0, 1000], [1000, 2000], [2000, 3000]])
        newshape = list(self.dat.data.shape)
        newshape[1] = 3
        self.assertEqual(list(dat.data.shape), newshape)
        # first epo (0)
        self.assertEqual(dat.data[0, 0, 0], 0)
        self.assertEqual(dat.data[0, 1, 0], 0)
        self.assertEqual(dat.data[0, 2, 0], 0)
        # second epo (1)
        self.assertEqual(dat.data[1, 0, 0], 1)
        self.assertEqual(dat.data[1, 1, 0], 2)
        self.assertEqual(dat.data[1, 2, 0], 3)
        # third epo (2)
        self.assertEqual(dat.data[2, 0, 0], 2)
        self.assertEqual(dat.data[2, 1, 0], 4)
        self.assertEqual(dat.data[2, 2, 0], 6)
        # fourth epo (0)
        self.assertEqual(dat.data[3, 0, 0], 0)
        self.assertEqual(dat.data[3, 1, 0], 0)
        self.assertEqual(dat.data[3, 2, 0], 0)
        # with one ival
        dat = jumping_means(self.dat, [[0, 1000]])
        newshape = list(self.dat.data.shape)
        newshape[1] = 1
        self.assertEqual(list(dat.data.shape), newshape)
        # first epo (0)
        self.assertEqual(dat.data[0, 0, 0], 0)
        # second epo (1)
        self.assertEqual(dat.data[1, 0, 0], 1)
        # third epo (2)
        self.assertEqual(dat.data[2, 0, 0], 2)
        # fourth epo (0)
        self.assertEqual(dat.data[3, 0, 0], 0)

    def test_jumping_means_with_cnt(self):
        """jumping_means must work with cnt argument."""
        data = self.dat.data[1]
        axes = self.dat.axes[1:]
        names = self.dat.names[1:]
        units = self.dat.units[1:]
        dat = self.dat.copy(data=data, axes=axes, names=names, units=units)
        dat = jumping_means(dat, [[0, 1000], [1000, 2000]])
        self.assertEqual(dat.data[0, 0], 1)
        self.assertEqual(dat.data[1, 0], 2)

    def test_jumping_means_swapaxes(self):
        """jumping means must work with nonstandard timeaxis."""
        dat = jumping_means(swapaxes(self.dat, 1, 2), [[0, 1000]], timeaxis=2)
        dat = swapaxes(dat, 1, 2)
        dat2 = jumping_means(self.dat, [[0, 1000]])
        self.assertEqual(dat, dat2)

    def test_jumping_means_copy(self):
        """jumping means must not modify argument."""
        cpy = self.dat.copy()
        jumping_means(self.dat, [[0, 1000]])
        self.assertEqual(self.dat, cpy)


if __name__ == '__main__':
    unittest.main()



from __future__ import division

import unittest

import numpy as np

from wyrm.types import Data
from wyrm.processing import select_epochs
from wyrm.processing import swapaxes


class TestSelectEpochs(unittest.TestCase):

    def setUp(self):
        ones = np.ones((10, 5))
        channels = ['ca1', 'ca2', 'cb1', 'cb2', 'cc1']
        time = np.linspace(0, 1000, 10, endpoint=False)
        classes = [0, 1, 2, 1]
        class_names = ['zeros', 'ones', 'twoes']
        # four cnts: 0s, 1s, -1s, and 0s
        data = np.array([ones * 0, ones * 1, ones * 2, ones * 0])
        self.dat = Data(data, [classes, time, channels], ['class', 'time', 'channel'], ['#', 'ms', '#'])
        self.dat.class_names = class_names

    def test_select_epochs(self):
        """Selecting Epochs."""
        # normal case
        dat = select_epochs(self.dat, [0])
        self.assertEqual(dat.data.shape[0], 1)
        np.testing.assert_array_equal(dat.data, self.dat.data[[0]])
        np.testing.assert_array_equal(dat.axes[0], self.dat.axes[0][0])
        # normal every second
        dat = select_epochs(self.dat, [0, 2])
        self.assertEqual(dat.data.shape[0], 2)
        np.testing.assert_array_equal(dat.data, self.dat.data[::2])
        np.testing.assert_array_equal(dat.axes[0], self.dat.axes[0][::2])
        # the full epo
        dat = select_epochs(self.dat, range(self.dat.data.shape[0]))
        np.testing.assert_array_equal(dat.data, self.dat.data)
        np.testing.assert_array_equal(dat.axes[0], self.dat.axes[0])
        # remove one
        dat = select_epochs(self.dat, [0], invert=True)
        self.assertEqual(dat.data.shape[0], 3)
        np.testing.assert_array_equal(dat.data, self.dat.data[1:])
        np.testing.assert_array_equal(dat.axes[0], self.dat.axes[0][1:])
        # remove every second
        dat = select_epochs(self.dat, [0, 2], invert=True)
        self.assertEqual(dat.data.shape[0], 2)
        np.testing.assert_array_equal(dat.data, self.dat.data[1::2])
        np.testing.assert_array_equal(dat.axes[0], self.dat.axes[0][1::2])

    def test_select_epochs_with_cnt(self):
        """Select epochs must raise an exception if called with cnt argument."""
        del(self.dat.class_names)
        with self.assertRaises(AssertionError):
            select_epochs(self.dat, [0, 1])

    def test_select_epochs_swapaxes(self):
        """Select epochs must work with nonstandard classaxis."""
        dat = select_epochs(swapaxes(self.dat, 0, 2), [0, 1], classaxis=2)
        dat = swapaxes(dat, 0, 2)
        dat2 = select_epochs(self.dat, [0, 1])
        self.assertEqual(dat, dat2)

    def test_select_epochs_copy(self):
        """Select Epochs must not modify argument."""
        cpy = self.dat.copy()
        select_epochs(self.dat, [0, 1])
        self.assertEqual(self.dat, cpy)


if __name__ == '__main__':
    unittest.main()



from __future__ import division

import unittest

import numpy as np

from wyrm.types import Data
from wyrm.processing import rectify_channels

class TestRectifytChannels(unittest.TestCase):

    def setUp(self):
        raw = np.arange(20).reshape(4, 5)
        channels = ['ca1', 'ca2', 'cb1', 'cb2', 'cc1']
        time = np.arange(4)
        self.dat = Data(raw, [time, channels], ['time', 'channels'], ['ms', '#'])

    def test_rectify_channels(self):
        """Rectify channels of positive and negative data must be equal."""
        dat = rectify_channels(self.dat.copy(data=-self.dat.data))
        dat2 = rectify_channels(self.dat)
        self.assertEqual(dat, dat2)

    def test_rectify_channels_copy(self):
        """Rectify channels must not change the original parameter."""
        cpy = self.dat.copy()
        rectify_channels(self.dat)
        self.assertEqual(cpy, self.dat)

if __name__ == '__main__':
    unittest.main()



from __future__ import division

import unittest

import numpy as np

from wyrm.types import Data, BlockBuffer
from wyrm.processing import append_cnt


class TestBlockBuffer(unittest.TestCase):

    def setUp(self):
        self.empty_dat = Data(np.array([]), [], [], [])
        data = np.array([0, 0])
        self.dat_1 = Data(np.array([0, 0])[np.newaxis, :], [np.array([0]), np.array(['ch1', 'ch2'])], ['time', 'channel'], ['ms', '#'])
        self.dat_1.fs = 1000
        self.dat_1.markers = [[0, 'x']]
        self.dat_5 = reduce(append_cnt, [self.dat_1 for i in range(5)])

    def test_append_empty(self):
        """Appending several emtpy dats must not modify the Block Buffer."""
        b = BlockBuffer(5)
        b.append(self.empty_dat)
        b.append(self.empty_dat)
        b.append(self.empty_dat)
        b.append(self.empty_dat)
        b.append(self.empty_dat)
        b.append(self.empty_dat)
        self.assertEqual(self.empty_dat, b.get())

    def test_append_until_full(self):
        """Appending fractions of block_length, must accumulate in the buffer until block_length is reached."""
        b = BlockBuffer(5)
        for i in range(4):
            b.append(self.dat_1)
        ret = b.get()
        self.assertEqual(self.empty_dat, ret)
        b.append(self.dat_1)
        ret = b.get()
        self.assertEqual(self.dat_5, ret)

    def test_append_with_markers(self):
        """Check if markers are handled correctly."""
        markers = [[i, 'x'] for i in range(5)]
        b = BlockBuffer(5)
        for i in range(4):
            b.append(self.dat_1)
        ret = b.get()
        self.assertEqual(self.empty_dat, ret)
        b.append(self.dat_1)
        ret = b.get()
        self.assertEqual(ret.markers, markers)


if __name__ == '__main__':
    unittest.main()



from __future__ import division

import unittest

import numpy as np

from wyrm.types import Data
from wyrm.processing import select_ival
from wyrm.processing import swapaxes


class TestSelectIval(unittest.TestCase):

    def setUp(self):
        ones = np.ones((10, 5))
        channels = ['ca1', 'ca2', 'cb1', 'cb2', 'cc1']
        time = np.linspace(-1000, 0, 10, endpoint=False)
        classes = [0, 0, 0]
        # three cnts: 1s, -1s, and 0s
        data = np.array([ones, ones * -1, ones * 0])
        self.dat = Data(data, [classes, time, channels], ['class', 'time', 'channel'], ['#', 'ms', '#'])
        self.dat.fs = 10

    def test_select_ival(self):
        """Selecting Intervals."""
        # normal case
        dat = select_ival(self.dat, [-500, 0])
        self.assertEqual(dat.axes[1][0], -500)
        self.assertEqual(dat.axes[1][-1],-100)
        # the full dat interval
        dat = select_ival(self.dat, [self.dat.axes[1][0], self.dat.axes[1][-1] + 1])
        self.assertEqual(dat.axes[1][0], self.dat.axes[1][0])
        self.assertEqual(dat.axes[1][-1], self.dat.axes[1][-1])
        np.testing.assert_array_equal(dat.data, self.dat.data)

    def test_select_ival_with_markers(self):
        """Selecting Intervals with markers."""
        # normal case
        good_markers = [[-499,99, 'x'], [-500, 'x'], [-0.0001, 'x']]
        bad_markers = [[501, 'y'], [0, 'y'], [1, 'y']]
        self.dat.markers = good_markers[:]
        self.dat.markers.extend(bad_markers)
        dat = select_ival(self.dat, [-500, 0])
        self.assertEqual(dat.markers, good_markers)

    def test_ival_checks(self):
        """Test for malformed ival parameter."""
        with self.assertRaises(AssertionError):
            select_ival(self.dat, [0, -1])
        with self.assertRaises(AssertionError):
            select_ival(self.dat, [self.dat.axes[1][0]-1, 0])
        with self.assertRaises(AssertionError):
            select_ival(self.dat, [0, self.dat.axes[1][-1]+1])

    def test_select_ival_copy(self):
        """Select_ival must not modify the argument."""
        cpy = self.dat.copy()
        select_ival(cpy, [-500, 0])
        self.assertEqual(cpy, self.dat)

    def test_select_ival_swapaxes(self):
        """select_ival must work with nonstandard timeaxis."""
        dat = select_ival(swapaxes(self.dat, 0, 1), [-500, 0], timeaxis=0)
        dat = swapaxes(dat, 0, 1)
        dat2 = select_ival(self.dat, [-500, 0])
        self.assertEqual(dat, dat2)

if __name__ == '__main__':
    unittest.main()



from __future__ import division

import unittest

import numpy as np

from wyrm.types import Data
from wyrm.processing import append
from wyrm.processing import swapaxes


class TestAppend(unittest.TestCase):

    def setUp(self):
        ones = np.ones((10, 5))
        # cnt with 1, 2, 3
        cnt = np.append(ones, ones*2, axis=0)
        cnt = np.append(cnt, ones*3, axis=0)
        channels = ['ca1', 'ca2', 'cb1', 'cb2', 'cc1']
        time = np.linspace(0, 3000, 30, endpoint=False)
        classes = [0, 1, 2, 1]
        # four cnts: 1s, -1s, and 0s
        data = np.array([cnt * 0, cnt * 1, cnt * 2, cnt * 0])
        self.dat = Data(data, [classes, time, channels], ['class', 'time', 'channel'], ['#', 'ms', '#'])

    def test_append(self):
        """Append."""
        dat = append(self.dat, self.dat)
        self.assertEqual(dat.data.shape[0], 2*self.dat.data.shape[0])
        self.assertEqual(len(dat.axes[0]), 2*len(self.dat.axes[0]))
        np.testing.assert_array_equal(dat.data, np.concatenate([self.dat.data, self.dat.data], axis=0))
        np.testing.assert_array_equal(dat.axes[0], np.concatenate([self.dat.axes[0], self.dat.axes[0]]))

    def test_append_with_extra(self):
        """append with extra must work with list and ndarrays."""
        self.dat.a = range(10)
        self.dat.b = np.arange(10)
        dat = append(self.dat, self.dat, extra=['a', 'b'])
        self.assertEqual(dat.a, range(10) + range(10))
        np.testing.assert_array_equal(dat.b, np.concatenate([np.arange(10), np.arange(10)]))

    def test_append_with_different_extra_types(self):
        """append must throw a TypeError if extra-types don't match."""
        a = self.dat.copy()
        b = self.dat.copy()
        a.a = range(10)
        b.a = np.arange(10)
        with self.assertRaises(TypeError):
            append(a, b, extra=['a'])

    def test_append_with_wrong_known_attributes(self):
        # .data dimensions must match
        a = self.dat.copy()
        a.data = a.data[np.newaxis, ...]
        with self.assertRaises(AssertionError):
            append(self.dat, a)
        # .data.shape must match for all axes except the ones beeing
        # appended
        a = self.dat.copy()
        a.data = a.data[..., :-1]
        with self.assertRaises(AssertionError):
            append(self.dat, a)
        # .axes must be equal for all axes except the ones beeing
        # appended
        a = self.dat.copy()
        a.axes[-1][0] = 'foo'
        with self.assertRaises(AssertionError):
            append(self.dat, a)
        # names must be equal
        a = self.dat.copy()
        a.names[0] = 'foo'
        with self.assertRaises(AssertionError):
            append(self.dat, a)
        # units must be equal
        a = self.dat.copy()
        a.units[0] = 'foo'
        with self.assertRaises(AssertionError):
            append(self.dat, a)

    def test_append_with_unsupported_extra_types(self):
        """append must trhow a TypeError if extra-type is unsupported."""
        self.dat.a = {'foo' : 'bar'}
        with self.assertRaises(TypeError):
            append(self.dat, self.dat, extra=['a'])

    def test_append_with_cnt(self):
        """append must work with cnt argument."""
        data = self.dat.data[1]
        axes = self.dat.axes[1:]
        names = self.dat.names[1:]
        units = self.dat.units[1:]
        dat = self.dat.copy(data=data, axes=axes, names=names, units=units)
        dat2 = append(dat, dat)
        self.assertEqual(dat2.data.shape[0], 2*dat.data.shape[0])
        self.assertEqual(len(dat2.axes[0]), 2*len(dat.axes[0]))
        np.testing.assert_array_equal(dat2.data, np.concatenate([dat.data, dat.data], axis=0))
        np.testing.assert_array_equal(dat2.axes[0], np.concatenate([dat.axes[0], dat.axes[0]], axis=0))

    def test_append_with_negative_axis(self):
        """Append must work correctly with a negative axis."""
        dat2 = self.dat.copy()
        dat2.data = dat2.data[:-1, ...]
        dat2.axes[0] = dat2.axes[2][:-1]
        a = append(self.dat, dat2, axis=0)
        b = append(self.dat, dat2, axis=-3)
        self.assertEqual(a, b)

    def test_append_swapaxes(self):
        """append must work with nonstandard timeaxis."""
        dat = append(swapaxes(self.dat, 0, 2), swapaxes(self.dat, 0, 2), axis=2)
        dat = swapaxes(dat, 0, 2)
        dat2 = append(self.dat, self.dat)
        self.assertEqual(dat, dat2)

    def test_append_copy(self):
        """append means must not modify argument."""
        cpy = self.dat.copy()
        append(self.dat, self.dat)
        self.assertEqual(self.dat, cpy)


if __name__ == '__main__':
    unittest.main()



from __future__ import division

import unittest

import numpy as np

from wyrm.types import Data
from wyrm.processing import clear_markers
from wyrm.processing import swapaxes


class TestClearMarkers(unittest.TestCase):

    def setUp(self):
        ones = np.ones((10, 5))
        # cnt with 1, 2, 3
        cnt = np.append(ones, ones*2, axis=0)
        cnt = np.append(cnt, ones*3, axis=0)
        channels = ['ca1', 'ca2', 'cb1', 'cb2', 'cc1']
        time = np.linspace(-1000, 2000, 30, endpoint=False)
        self.good_markers = [[-1000, 'a'], [-999, 'b'], [0, 'c'], [1999.9999999, 'd']]
        bad_markers = [[-1001, 'x'], [2000, 'x']]
        markers = self.good_markers[:]
        markers.extend(bad_markers)
        classes = [0, 1, 2, 1]
        # four cnts: 1s, -1s, and 0s
        data = np.array([cnt * 0, cnt * 1, cnt * 2, cnt * 0])
        self.dat = Data(data, [classes, time, channels], ['class', 'time', 'channel'], ['#', 'ms', '#'])
        self.dat.markers = markers
        self.dat.fs = 10

    def test_clear_markers(self):
        """Clear markers."""
        dat = clear_markers(self.dat)
        self.assertEqual(dat.markers, self.good_markers)

    def test_clear_emtpy_markers(self):
        """Clearing emtpy markers has no effect."""
        dat = self.dat.copy()
        dat.markers = []
        dat2 = clear_markers(dat)
        self.assertEqual(dat, dat2)

    def test_clear_nonexisting_markers(self):
        """Clearing emtpy markers has no effect."""
        dat = self.dat.copy()
        del dat.markers
        dat2 = clear_markers(dat)
        self.assertEqual(dat, dat2)

    def test_clear_markers_w_empty_data(self):
        """Clearing emtpy dat should remove all markers."""
        dat = self.dat.copy()
        dat.data = np.array([])
        dat2 = clear_markers(dat)
        self.assertEqual(dat2.markers, [])

    def test_clear_markes_swapaxes(self):
        """clear_markers must work with nonstandard timeaxis."""
        dat = clear_markers(swapaxes(self.dat, 1, 2), timeaxis=2)
        dat = swapaxes(dat, 1, 2)
        dat2 = clear_markers(self.dat)
        self.assertEqual(dat, dat2)

    def test_clear_markers_copy(self):
        """clear_markers must not modify argument."""
        cpy = self.dat.copy()
        clear_markers(self.dat)
        self.assertEqual(self.dat, cpy)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python


import unittest

import numpy as np

from wyrm.types import Data


class TestData(unittest.TestCase):

    def setUp(self):
        self.data = np.arange(20).reshape(4, 5)
        self.axes = [np.arange(4), np.arange(5)]
        self.names = ['foo', 'bar']
        self.units = ['u1', 'u2']

    def test_init(self):
        """Test init with correct values."""
        d = Data(self.data, self.axes, self.names, self.units)
        np.testing.assert_array_equal(d.data, self.data)
        for a, b in zip(d.axes, self.axes):
            np.testing.assert_array_equal(a, b)
        self.assertEqual(self.names, d.names)
        self.assertEqual(self.units, d.units)

    def test_init_emtpy(self):
        """Test init with empty data."""
        d = Data(np.array([]), [], [], [])
        self.assertFalse(d)

    def test_init_emtpy2(self):
        """With emtpy data it does not matter what the rest is."""
        d = Data(np.array([]), [1, 2, 3], [1, 1, 2], [1, 2, 3])
        self.assertFalse(d)

    def test_init_with_inconsistent_values(self):
        """Test init with inconsistent values."""
        data = self.data[np.newaxis, :]
        with self.assertRaises(AssertionError):
            Data(data, self.axes, self.names, self.units)
        axes = self.axes[:]
        axes[0] = np.arange(100)
        with self.assertRaises(AssertionError):
            Data(self.data, axes, self.names, self.units)
        names = self.names[:]
        names.append('baz')
        with self.assertRaises(AssertionError):
            Data(self.data, self.axes, names, self.units)
        units = self.units[:]
        units.append('u3')
        with self.assertRaises(AssertionError):
            Data(self.data, self.axes, self.names, units)

    def test_truth_value(self):
        """Test __nonzero__."""
        d = Data(np.array([]), [], [], [])
        self.assertFalse(d)
        d = Data(self.data, self.axes, self.names, self.units)
        self.assertTrue(d)

    def test_equality(self):
        """Test the various (in)equalities."""
        d1 = Data(self.data, self.axes, self.names, self.units)
        # known extra attributes
        d1.markers = [[123, 'foo'], [234, 'bar']]
        d1.fs = 100
        # unknown extra attribute
        d1.foo = 'bar'
        # so far, so equal
        d2 = d1.copy()
        self.assertEqual(d1, d2)
        # different shape
        d2 = d1.copy()
        d2.data = np.arange(20).reshape(5, 4)
        self.assertNotEqual(d1, d2)
        # different data
        d2 = d1.copy()
        d2.data[0, 0] = 42
        self.assertNotEqual(d1, d2)
        # different axes
        d2 = d1.copy()
        d2.axes[0] = np.arange(100)
        self.assertNotEqual(d1, d2)
        # different names
        d2 = d1.copy()
        d2.names[0] = 'baz'
        self.assertNotEqual(d1, d2)
        # different untis
        d2 = d1.copy()
        d2.units[0] = 'u3'
        self.assertNotEqual(d1, d2)
        # different known extra attribute
        d2 = d1.copy()
        d2.markers[0] = [123, 'baz']
        self.assertNotEqual(d1, d2)
        # different known extra attribute
        d2 = d1.copy()
        d2.fs = 10
        self.assertNotEqual(d1, d2)
        # different unknown extra attribute
        d2 = d1.copy()
        d2.baz = 'baz'
        self.assertNotEqual(d1, d2)
        # different new unknown extra attribute
        d2 = d1.copy()
        d2.bar = 42
        self.assertNotEqual(d1, d2)

    def test_eq_and_ne(self):
        """Check if __ne__ is properly implemented."""
        d1 = Data(self.data, self.axes, self.names, self.units)
        d2 = d1.copy()
        # if __eq__ is implemented and __ne__ is not, this evaluates to
        # True!
        self.assertFalse(d1 == d2 and d1 != d2)

    def test_copy(self):
        """Copy must work."""
        d1 = Data(self.data, self.axes, self.names, self.units)
        d2 = d1.copy()
        self.assertEqual(d1, d2)
        # we can't really check of all references to be different in
        # depth recursively, so we only check on the first level
        for k in d1.__dict__:
            self.assertNotEqual(id(getattr(d1, k)), id(getattr(d2, k)))
        d2 = d1.copy(foo='bar')
        self.assertEqual(d2.foo, 'bar')



if __name__ == '__main__':
    unittest.main()





"""Input/Output methods.

This module provides methods for loading and saving data into various
formats.

"""

from __future__ import division

from os import path
import logging
import re

import numpy as np

from wyrm.types import Data


logging.basicConfig(level=logging.NOTSET)
logger = logging.getLogger(__name__)



def load_brain_vision_data(vhdr):
    """Load Brain Vision data from a file.

    This methods loads the continuous EEG data, and returns a ``Data``
    object of continuous data ``[time, channel]``, along with the
    markers and the sampling frequency. The EEG data is returned in
    micro Volt.

    Parameters
    ----------
    vhdr : str
        Path to a VHDR file

    Returns
    -------
    dat : Data
        Continuous Data with the additional attributes ``.fs`` for the
        sampling frequency and ``.marker`` for a list of markers. Each
        marker is a tuple of ``(time in ms, marker)``.

    Raises
    ------
    AssertionError
        If one of the consistency checks fails

    Examples
    --------

    >>> dat = load_brain_vision_data('path/to/vhdr')
    >>> dat.fs
    1000
    >>> dat.data.shape
    (54628, 61)

    """
    logger.debug('Loading Brain Vision Data Exchange Header File')
    with open(vhdr) as fh:
        fdata = map(str.strip, fh.readlines())
    fdata = filter(lambda x: not x.startswith(';'), fdata)
    fdata = filter(lambda x: len(x) > 0, fdata)
    # check for the correct file version:
    assert fdata[0].endswith('1.0')
    # read all data into a dict where the key is the stanza of the file
    file_dict = dict()
    for line in fdata[1:]:
        if line.startswith('[') and line.endswith(']'):
            current_stanza = line[1:-1]
            file_dict[current_stanza] = []
        else:
            file_dict[current_stanza].append(line)
    # translate known stanzas from simple list of strings to a dict
    for stanza in 'Common Infos', 'Binary Infos', 'Channel Infos':
        logger.debug(stanza)
        file_dict[stanza] = {line.split('=', 1)[0]: line.split('=', 1)[1] for line in file_dict[stanza]}
    # now file_dict contains the parsed data from the vhdr file
    # load the rest
    data_f = file_dict['Common Infos']['DataFile']
    marker_f = file_dict['Common Infos']['MarkerFile']
    data_f = path.sep.join([path.dirname(vhdr), data_f])
    marker_f = path.sep.join([path.dirname(vhdr), marker_f])
    n_channels = int(file_dict['Common Infos']['NumberOfChannels'])
    sampling_interval_microseconds = float(file_dict['Common Infos']['SamplingInterval'])
    fs = 1 / (sampling_interval_microseconds / 10**6)
    channels = [file_dict['Channel Infos']['Ch%i' % (i + 1)] for i in range(n_channels)]
    channels = map(lambda x: x.split(',')[0], channels)
    resolutions = [file_dict['Channel Infos']['Ch%i' % (i + 1)] for i in range(n_channels)]
    resolutions = map(lambda x: float(x.split(',')[2]), resolutions)
    # assert all channels have the same resolution of 0.1
    # FIXME: that is not always true, for example if we measure pulse or
    # emg
    #assert all([i == 0.1 for i in resolutions])
    # some assumptions about the data...
    assert file_dict['Common Infos']['DataFormat'] == 'BINARY'
    assert file_dict['Common Infos']['DataOrientation'] == 'MULTIPLEXED'
    assert file_dict['Binary Infos']['BinaryFormat'] == 'INT_16'
    # load EEG data
    logger.debug('Loading EEG Data.')
    data = np.fromfile(data_f, np.int16)
    data = data.reshape(-1, n_channels)
    data *= resolutions[0]
    n_samples = data.shape[0]
    # duration in ms
    duration = 1000 * n_samples / fs
    time = np.linspace(0, duration, n_samples, endpoint=False)
    # load marker
    logger.debug('Loading Marker.')
    regexp = r'^Mk(?P<mrk_nr>[0-9]*)=.*,(?P<mrk_descr>.*),(?P<mrk_pos>[0-9]*),[0-9]*,[0-9]*$'
    mrk = []
    with open(marker_f) as fh:
        for line in fh:
            line = line.strip()
            match = re.match(regexp, line)
            if match is None:
                continue
            mrk_pos = match.group('mrk_pos')
            mrk_descr = match.group('mrk_descr')
            if len(mrk_descr) > 1:
                # marker := [samplenr, marker]
                #mrk.append([int(mrk_pos), mrk_descr])
                # marker := [time in ms, marker]
                mrk.append([time[int(mrk_pos)], mrk_descr])
    dat = Data(data, [time, channels], ['time', 'channel'], ['ms', '#'])
    dat.fs = fs
    dat.markers = mrk
    return dat





from __future__ import division

import unittest

import numpy as np
from scipy.fftpack import rfft, rfftfreq
from scipy.signal import butter

from wyrm.types import Data
from wyrm.processing import lfilter
from wyrm.processing import swapaxes


class TestLFilter(unittest.TestCase):

    def setUp(self):
        # create some data
        fs = 100
        dt = 5
        self.freqs = [2, 7, 15]
        amps = [30, 10, 2]
        t = np.linspace(0, dt, fs*dt)
        data = np.sum([a * np.sin(2*np.pi*t*f) for a, f in zip(amps, self.freqs)], axis=0)
        data = data[:, np.newaxis]
        data = np.concatenate([data, data], axis=1)
        channel = np.array(['ch1', 'ch2'])
        self.dat = Data(data, [t, channel], ['time', 'channel'], ['s', '#'])
        self.dat.fs = fs

    def test_bandpass(self):
        """Band pass filtering."""
        # bandpass around the middle frequency
        fn = self.dat.fs / 2
        b, a = butter(4, [6 / fn, 8 / fn], btype='band')
        ans = lfilter(self.dat, b, a)
        # the amplitudes
        fourier = np.abs(rfft(ans.data, axis=0) * 2 / self.dat.data.shape[0])
        ffreqs = rfftfreq(ans.data.shape[0], 1/ans.fs)
        # check if the outer freqs are damped close to zero
        # freqs...
        for i in self.freqs[0], self.freqs[-1]:
            # buckets for freqs
            for j in fourier[ffreqs == i]:
                # channels
                for k in j:
                    self.assertAlmostEqual(k, 0., delta=.1)

    def test_lfilter_copy(self):
        """lfilter must not modify argument."""
        cpy = self.dat.copy()
        fn = self.dat.fs / 2
        b, a = butter(4, [6 / fn, 8 / fn], btype='band')
        lfilter(self.dat, b, a)
        self.assertEqual(cpy, self.dat)

    def test_lfilter_swapaxes(self):
        """lfilter must work with nonstandard timeaxis."""
        fn = self.dat.fs / 2
        b, a = butter(4, [6 / fn, 8 / fn], btype='band')
        dat = lfilter(swapaxes(self.dat, 0, 1), b, a, timeaxis=1)
        dat = swapaxes(dat, 0, 1)
        dat2 = lfilter(self.dat, b, a)
        self.assertEqual(dat, dat2)


if __name__ == '__main__':
    unittest.main()
